/*
 * Copyright 2013 Akiban Technologies, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.akiban.jonquil;

import com.google.common.base.Function;
import com.google.common.base.Joiner;
import com.google.common.base.Strings;
import com.google.common.collect.Collections2;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

public abstract class EntityDescriber {

    public abstract Collection<? extends String> getCollections(String name);

    public void prettyPrint(PrintStream out, int indentSize) {
        List<Line> lines = new ArrayList<>();

        for (Map.Entry<String, Collection<String>> entry : ids.entrySet()) {
            String entryName = entry.getKey();
            if (roots.contains(entryName)) {
                prettyPrint(lines, indentSize, null, entryName, 0);
            }
        }
        int namesWidth = NAME_LABEL.length();
        int pksWidth = IDENTIFYING_FIELDS_LABEL.length();
        for (Line line : lines) {
            namesWidth = Math.max(namesWidth, line.collection.length());
            pksWidth = Math.max(pksWidth, line.pkDescs.length());
        }
        namesWidth += indentSize;
        pksWidth += indentSize;
        String formatString = "%-" + namesWidth + 's' + "%-" + pksWidth + "s%s%n";
        out.printf(formatString, NAME_LABEL, IDENTIFYING_FIELDS_LABEL, PARENT_REFS_LABEL);
        out.println(Strings.repeat("-", namesWidth+pksWidth+PARENT_REFS_LABEL.length()));
        for (Line line : lines) {
            String fkString = (line.fksDescs == null)
                    ? ""
                    : joiner.join(line.fksDescs);
            out.printf(formatString, line.collection, line.pkDescs, fkString);
        }
    }

    protected EntityDescriber(Map<String, Collection<String>> ids, Set<String> roots, JoinStrategy joins) {
        this.ids = ids;
        this.roots = roots;
        this.joins = joins;
    }

    private final Map<String, Collection<String>> ids;
    private final Set<String> roots;
    private final JoinStrategy joins;

    private static final Joiner joiner = Joiner.on(", ");
    private static final String NAME_LABEL = "name";
    private static final String IDENTIFYING_FIELDS_LABEL = "identifying fields";
    private static final String PARENT_REFS_LABEL = "parent refs";

    private void prettyPrint(List<Line> out, int indentSize, String parentName, String name, int initialIndent) {
        out.add(new Line(initialIndent, name, ids.get(name), parentName));
        int nextIndent = initialIndent + indentSize;
        for (String collection : getCollections(name))
            prettyPrint(out, indentSize, name, collection, nextIndent);
    }

    private static final class JoinFieldsFormatter implements Function<JoinFields, String> {

        @Override
        public String apply(JoinFields input) {
            return String.format("%s -> %s.%s", input.getChildField(), parent, input.getParentField());
        }

        private JoinFieldsFormatter(String parent) {
            this.parent = parent;
        }

        private final String parent;
    }

    private final class Line {
        public Line(int indentSize, String collection, Collection<String> ids, String parentName) {
            this.collection = String.format("%s%s", Strings.repeat(" ", indentSize), collection);
            this.pkDescs = joiner.join(ids);
            Collection<? extends JoinFields> fks = joins.getJoins(parentName, collection);
            fksDescs = (fks == null)
                    ? null
                    : Collections2.transform(fks, new JoinFieldsFormatter(parentName));
        }

        private final String collection;
        private final String pkDescs;
        private final Collection<? extends String> fksDescs;
    }
}
