/*
 * Copyright 2013 Akiban Technologies, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.akiban.jonquil;

import com.akiban.jonquil.actions.Action;
import com.akiban.jonquil.actions.Actions;
import com.google.common.base.Function;
import com.google.common.base.Functions;
import com.google.common.base.Joiner;
import org.codehaus.jackson.JsonNode;
import org.codehaus.jackson.JsonParser;
import org.codehaus.jackson.JsonToken;

import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

public final class Jonquil {

    public static String createSQL(String rootTable, JsonParser parser, JoinStrategy joinStrategy) throws IOException {
        return createSQL(rootTable, parser, joinStrategy, Functions.constant((Action)null));
    }

    public static String createSQL(String rootTable, JsonParser parser, JoinStrategy joinStrategy,
                                   Function<? super String, ? extends Action> additionalActions) throws IOException
    {
        Jonquil fragment = new Jonquil(parser, joinStrategy, additionalActions);
        try {
            if (fragment.input.nextToken() != JsonToken.START_OBJECT)
                throw new JonquilException("json document must be an object");
            JonquilWriter writer = fragment.startSelect(rootTable);
            return writer.finish();
        }
        catch (JonquilException e) {
            e.setLocation(fragment.input.getCurrentLocation());
            throw e;
        }
        catch (Exception e) {
            JonquilException ae = new JonquilException(e.getMessage(), e);
            ae.setLocation(fragment.input.getCurrentLocation());
            throw ae;
        }
    }

    public static String escapeName(String text) {
        return '"' + text.replace("\"", "\"\"") + '"';
    }

    public static String escapeValue(String text) {
        return '\'' + text.replace("'", "''") + '\'';
    }

    private JonquilWriter startSelect(String rootTable) throws IOException {
        JonquilWriter writer = new JonquilWriter(rootTable);
        expectCurrent(JsonToken.START_OBJECT);
        while (nextIsnt(JsonToken.END_OBJECT)) {
            expectCurrent(JsonToken.FIELD_NAME);
            String fieldName = input.getCurrentName();
            if (fieldName.isEmpty())
                throw new JonquilException("null or empty field names are not allowed");
            else if (fieldName.charAt(0) == '@') {
                String actionName = fieldName.substring(1);
                Action action = additionalActions.apply(actionName);
                if (action == null)
                    action = Actions.findAction(actionName);
                action.apply(input, writer, rootTable);
            }
            else {
                switch (input.nextToken()) {
                case VALUE_NULL:
                    throw new JonquilException("values can't be null");
                case START_ARRAY:
                    arrayValue(writer, fieldName);
                    break;
                case VALUE_STRING:
                    stringValue(writer, fieldName);
                    break;
                case VALUE_NUMBER_INT:
                case VALUE_NUMBER_FLOAT:
                case VALUE_TRUE:
                case VALUE_FALSE:
                    nonStringScalarValue(writer, fieldName);
                    break;
                case START_OBJECT:
                    subSelectValue(writer, rootTable, fieldName);
                    break;
                default:
                    throw new JonquilException("unsupported token: " + input.getCurrentToken());
                }
            }
        }
        return writer;
    }

    private void subSelectValue(JonquilWriter writer, String parentTable, String childTable) throws IOException {
        JonquilWriter subSelect = startSelect(childTable);
        for (JoinFields join : joinStrategy.getJoins(parentTable, childTable)) {
            subSelect.addCondition(String.format(
                    "%s.%s = %s.%s",
                    escapeName(parentTable),
                    escapeName(join.getParentField()),
                    escapeName(childTable),
                    escapeName(join.getChildField())));
        }
        writer.addExpression(subSelect.finish(), childTable);
    }

    private void nonStringScalarValue(JonquilWriter writer, String fieldName) throws IOException {
        writer.addScalar(fieldName);
        addEquality(writer, fieldName, input.getText(), false);
    }

    private void stringValue(JonquilWriter writer, String fieldName) throws IOException {
        String fieldValue = input.getText();
        if (!fieldValue.isEmpty()) {
            char first = fieldValue.charAt(0);
            if (first == '=') {
                String expression = fieldValue.substring(1);
                if (expression.isEmpty())
                    throw new JonquilException(fieldName + ": can't have empty expression");
                writer.addExpression(expression, fieldName);
                return;
            }
        }
        writer.addScalar(fieldName);
        addEquality(writer, fieldName, fieldValue, true);
    }

    private void arrayValue(JonquilWriter writer, String fieldName) throws IOException {
        writer.addScalar(fieldName);
        List<String> values = new ArrayList<>();
        while (nextIsnt(JsonToken.END_ARRAY)) {
            switch (input.getCurrentToken()) {
            case VALUE_STRING:
                values.add(escapeValue(input.getText()));
                break;
            case VALUE_NUMBER_INT:
            case VALUE_NUMBER_FLOAT:
            case VALUE_TRUE:
            case VALUE_FALSE:
                values.add(input.getText());
                break;
            default:
                throw new JonquilException("invalid value in OR: " + input.getCurrentToken() + ' ' + input.getText());
            }
        }
        if (!values.isEmpty())
            writer.addCondition(escapeName(fieldName) + " IN(" + Joiner.on(',').join(values) +')');
    }

    private void addEquality(JonquilWriter wrier, String fieldName, String value, boolean escapeValue) {
        String rhs = escapeValue ? escapeValue(value) : value;
        wrier.addCondition(escapeName(fieldName) + "=" + rhs);
    }

    protected void expectCurrent(JsonToken expected) throws IOException {
        JsonToken actual = input.getCurrentToken();
        if (actual != expected)
            throw new JonquilException("expected " + expected + " but found " + actual);
    }

    protected boolean nextIsnt(JsonToken token) throws IOException {
        JsonToken next = input.nextToken();
        if (next == null)
            throw new JonquilException("unexpected end of input");
        return next != token;
    }

    public Jonquil(JsonParser input, JoinStrategy joinStrategy,
                 Function<? super String, ? extends Action> additionalActions)
    {
        this.input = input;
        this.joinStrategy = joinStrategy;
        this.additionalActions = additionalActions;
    }

    public static void main(String[] args) throws IOException {
        JsonNode sampleJoinsNode;
        try (InputStream sampleJoinsIS = Jonquil.class.getResourceAsStream("sample_joins.json")) {
            sampleJoinsNode = JonquilUtils.jsonFactory.createJsonParser(sampleJoinsIS).readValueAsTree();
        }
        JoinStrategy joinStrategy = JonquilUtils.jsonBasedJoinStrategy(sampleJoinsNode);
        JonquilUtils.jsonBasedEntityDescriber(joinStrategy, sampleJoinsNode).prettyPrint(System.out, 4);
        while(true) {
            System.out.println();
            System.out.println("Jonquil query:");
            try {
                JsonParser stdinParser = JonquilUtils.jsonFactory.createJsonParser(System.in);
                String sql = createSQL("customers", stdinParser, joinStrategy);
                System.out.println(sql + ';');
            }
            catch (JonquilException e) {
                System.err.println(e.getMessage());
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    private final JoinStrategy joinStrategy;
    private final JsonParser input;
    private final Function<? super String, ? extends Action> additionalActions;

    private static class NoCloseOutStream extends FilterOutputStream {
        private NoCloseOutStream(OutputStream out) {
            super(out);
        }

        @Override
        public void close() {}
    }
}
