/*
 * Copyright 2013 Akiban Technologies, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.akiban.jonquil;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.ListMultimap;
import com.google.common.collect.SortedSetMultimap;
import com.google.common.collect.TreeMultimap;
import org.codehaus.jackson.JsonFactory;
import org.codehaus.jackson.JsonNode;
import org.codehaus.jackson.map.ObjectMapper;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

public final class JonquilUtils {

    public static final ObjectMapper jsonMapper = new ObjectMapper();
    public static final JsonFactory jsonFactory = new JsonFactory(jsonMapper);

    public static JoinStrategy jsonBasedJoinStrategy(JsonNode node) {
        JoinStrategyBuilder builder = new JoinStrategyBuilder();
        buildJoins(builder, null, null, node);
        return builder.get();
    }

    public static EntityDescriber jsonBasedEntityDescriber(JoinStrategy joins, JsonNode node) {
        final ListMultimap<String,String> identifiers = ArrayListMultimap.create();
        final SortedSetMultimap<String, String> children = TreeMultimap.create();
        final Set<String> rootNames = new HashSet<>();

        for (Iterator<Map.Entry<String, JsonNode>> iter = node.getFields(); iter.hasNext(); ) {
            Map.Entry<String, JsonNode> root = iter.next();
            String rootName = root.getKey();
            rootNames.add(rootName);
            buildDescriberLevel(rootName, root.getValue(), identifiers, children);
        }
        return new EntityDescriber(identifiers.asMap(), rootNames, joins) {
            @Override
            public Collection<? extends String> getCollections(String name) {
                return Collections.unmodifiableSet(children.get(name));
            }
        };
    }

    private static void buildDescriberLevel(String name, JsonNode node,
                                            ListMultimap<String, String> identifiers,
                                            SortedSetMultimap<String, String> children)
    {
        if (!node.isObject())
            throw new IllegalArgumentException(name + " is not a json object");
        if (identifiers.containsKey(name) || children.containsKey(name))
            throw new IllegalArgumentException("entity names must be unique: " + name);

        JsonNode identifiersArray = node.get("identifiers");
        if (identifiersArray == null)
            throw new IllegalArgumentException("no identifiers provided for " + name);
        int nIdentifiers = identifiersArray.size();
        if (nIdentifiers <= 0)
            throw new IllegalArgumentException("\"identifiers\" must be a non-empty array: " + name);
        for (int i = 0; i < nIdentifiers; ++i) {
            JsonNode id = identifiersArray.get(i);
            if (!id.isTextual())
                throw new IllegalArgumentException("identifier must be textual: " + id + " for " + name);
            identifiers.put(name, id.getTextValue());
        }

        JsonNode childrenNode = node.get("children");
        if (childrenNode != null) {
            for (Iterator<Map.Entry<String, JsonNode>> childIter = childrenNode.getFields(); childIter.hasNext(); ) {
                Map.Entry<String, JsonNode> child = childIter.next();
                String childName = child.getKey();
                children.put(name, childName);
                buildDescriberLevel(childName, child.getValue(), identifiers, children);
            }
        }
    }

    private static void buildJoins(JoinStrategyBuilder builder, String parent, List<String> parentIds, JsonNode node) {
        for (Iterator<Map.Entry<String, JsonNode>> fields = node.getFields(); fields.hasNext(); ) {
            Map.Entry<String, JsonNode> joinLevel = fields.next();
            String tableName = joinLevel.getKey();
            JsonNode tableInfo = joinLevel.getValue();
            JsonNode toParentFields = tableInfo.get("to_parent");
            if (toParentFields != null) {
                assert parentIds != null;
                int len = toParentFields.size();
                assert parentIds.size() == len : toParentFields + " has different size than " + parentIds;
                for (int i = 0; i < len; ++i) {
                    String parentId = parentIds.get(i);
                    String toParent = toParentFields.get(i).getTextValue();
                    builder.addJoin(parent, parentId, tableName, toParent);
                }
            }
            JsonNode children = tableInfo.get("children");
            if (children != null) {
                // assume we have identifiers
                JsonNode idsArray = tableInfo.get("identifiers");
                int idsLen = idsArray.size();
                List<String> ids = new ArrayList<>(idsLen);
                for (int i = 0; i < idsLen; ++i) {
                    ids.add(idsArray.get(i).getTextValue());
                }
                buildJoins(builder, tableName, ids, children);
            }
        }
    }

    private JonquilUtils() {}
}
