/*
 * Copyright 2013 Akiban Technologies, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.akiban.jonquil;

import com.google.common.base.Joiner;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public final class JonquilWriter {

    public String finish() {
        if (scalars.isEmpty())
            throw new JonquilException("nothing to select!");
        StringBuilder sb = new StringBuilder();
        sb.append("SELECT ");
        List<String> scalarFragments = new ArrayList<>();
        for (Map.Entry<String,String> entry : scalars.entrySet()) {
            String as = Jonquil.escapeName(entry.getKey());
            String expr = entry.getValue();
            String fragment = (expr == null) ? as : String.format("(%s) %s", expr, as);
            scalarFragments.add(fragment);
        }
        sb.append(Joiner.on(", ").join(scalarFragments));
        sb.append(" FROM ").append(Jonquil.escapeName(tableName));
        if (!conditions.isEmpty()) {
            sb.append(" WHERE ");
            sb.append(Joiner.on(" AND ").join(conditions));
        }
        if (hasOrdering()) {
            assert ! orderings.isEmpty();
            sb.append(" ORDER BY ");
            for(Iterator<Map.Entry<String,Ordering>> iter = orderings.entrySet().iterator(); iter.hasNext(); ) {
                Map.Entry<String, Ordering> entry = iter.next();
                String field = Jonquil.escapeName(entry.getKey());
                sb.append(field);
                if (entry.getValue() != Ordering.ASC)
                    sb.append(' ').append(entry.getValue().name());

                if (iter.hasNext())
                    sb.append(", ");
            }
        }
        if (hasLimit())
            sb.append(" LIMIT ").append(limit);
        return sb.toString();
    }

    public void addCondition(String condition) {
        conditions.add(condition);
    }

    public void addScalar(String scalar) {
        if (!scalars.containsKey(scalar))
            scalars.put(scalar, null);
    }

    public void addExpression(String expression, String as) {
        String old = scalars.get(as);
        if ((old != null) && (!old.equals(expression)))
            throw new JonquilException("multiple expressions found for scalar name \"" + as + '"');
        scalars.put(as, expression);
    }

    public void setLimit(int limit) {
        if (hasLimit())
            throw new JonquilException("can't set multiple limits");
        this.limit = limit;
    }

    public void addOrderBy(String field, Ordering ordering) {
        if (!hasOrdering())
            orderings = new LinkedHashMap<>();
        if (orderings.containsKey(field))
            throw new JonquilException("already have ordering for field: " + field);
        orderings.put(field, ordering);
    }

    public JonquilWriter(String tableName) {
        this.tableName = tableName;
    }

    private final String tableName;
    private final LinkedHashMap<String,String> scalars = new LinkedHashMap<>();
    private final List<String> conditions = new ArrayList<>();
    private long limit = -1;
    private LinkedHashMap<String,Ordering> orderings;

    private boolean hasLimit() {
        return limit >= 0;
    }

    private boolean hasOrdering() {
        return orderings != null;
    }

    public enum Ordering {
        ASC, DESC;

        public static Ordering parse(String string) {
            switch (string.toLowerCase()) {
            case "asc":
            case "ascending":
                return ASC;
            case "desc":
            case "descending":
                return DESC;
            default:
                throw new JonquilException("ordering must be ASC or DESC");
            }
        }
    }
}
